package at.jku.fim.datalinksimulation.genericdevices;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import at.jku.fim.datalinksimulation.layer1.data.RawData;
import at.jku.fim.datalinksimulation.layer1.wire.Wire;
import at.jku.fim.datalinksimulation.layer1.wire.WireConnector;
import at.jku.fim.datalinksimulation.layer1.wire.WiringException;
import at.jku.fim.datalinksimulation.layer7.networkdevice.DataReflectApplication;
import at.jku.fim.datalinksimulation.networkstack.NetworkStack;
import at.jku.fim.datalinksimulation.networkstack.NetworkStackParameters;
import at.jku.fim.datalinksimulation.networkstack.incoming.IncomingNetworkStackConfiguration;
import at.jku.fim.datalinksimulation.networkstack.outgoing.OutgoingNetworkStackConfiguration;
import at.jku.fim.datalinksimulation.networkstack.outgoing.OutgoingNetworkStackObserver;
import at.jku.fim.datalinksimulation.utils.Logger;

/**
 * <p>Title: OSI Simulation Framework</p>
 * <p>Copyright: (c) 2004-05</p>
 * <p>Company: FIM@JKU, www.fim.uni-linz.ac.at</p>
 * @author Andreas Putzinger
 * @version $Id: InterlinkingDevice.java,v 1.1 2005/10/27 13:10:07 aputzinger Exp $
 */
public abstract class InterlinkingDevice extends HardwareComponent implements WireConnector {

	/** each interlinking device has its own log */
    protected final Logger logger;
	
    /** each interlinking device has it's own network stack */ 
    protected final NetworkStack networkStack;

    /** the number of available ports */
    protected final int numberOfPorts;
    
    /** reference to all connected wires, indexed by port number */
    private Wire[] connectedWires;

    /** internal identifier */
    private String identifier;
    
    /**
     * Constructor
     * @param identifier Identifier for log
     * @param numberOfPorts Number of ports of device
     */
    public InterlinkingDevice(String identifier, final int numberOfPorts) {
		connectedWires = new Wire[numberOfPorts];

		logger = new Logger(identifier);

		this.identifier = identifier;
		
		networkStack = new NetworkStack(
				getIncomingNetworkStackConfiguration(),
				getOutgoingNetworkStackConfiguration(), 
				identifier);
		
		networkStack.subscribeApplication(new DataReflectApplication(networkStack));

		networkStack.connectPhysical(new OutgoingNetworkStackObserver() {

			/*
			 * (non-Javadoc)
			 * 
			 * @see at.jku.fim.datalinksimulation.networkstack.OutgoingNetworkStackObserver#informAboutNewData(at.jku.fim.datalinksimulation.data.RawData,
			 *      at.jku.fim.datalinksimulation.networkstack.NetworkStackParameters)
			 */
			public void informAboutNewData(RawData data, NetworkStackParameters params) {
				List receiverPorts = null;
				if (params.getParameter(
						1,
						NetworkStack.NETWORK_PARAM_OUTGOING_PORT_NUMS) != null) {
					receiverPorts = (List)params.getParameter(
							1,
							NetworkStack.NETWORK_PARAM_OUTGOING_PORT_NUMS);				
				} else {
					int senderPort = -1;
					receiverPorts = new ArrayList();
					
					// determine sender port
					if (params.getParameter(
							1, NetworkStack.NETWORK_PARAM_INCOMING_PORT_NUM) != null) {
						senderPort = ((Integer)params.getParameter(
								1, NetworkStack.NETWORK_PARAM_INCOMING_PORT_NUM)).intValue();
					}
					
					// receiver ports = all ports except sender port
					for (int i = 0; i < numberOfPorts; i++) {
						if (getWireInPort(i) != null && i != senderPort) {
							receiverPorts.add(new Integer(i));
						}
					}
				}

				for (Iterator it = receiverPorts.iterator(); it.hasNext(); ) {
					int portNr = ((Integer)it.next()).intValue();
					if (getWireInPort(portNr) != null) {
						logger.log("Sending data out on port <" + portNr + ">");
						getWireInPort(portNr).send(InterlinkingDevice.this, data);
					}
				}
			}

		});
		
		networkStack.setParentLogger(logger);
		
		this.numberOfPorts = numberOfPorts;
		
	}

    /*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.layer1.wire.WireConnector#getNumberOfTotalPorts()
	 */
    public final int getNumberOfTotalPorts() {
    	return numberOfPorts;
    }

    /**
     * @return Returns the number of free and therefore available ports
     */
    public final int getNumberOfFreePorts() {
    	int result = 0;
    	for (int i = 0; i < getNumberOfTotalPorts(); i++) {
    		if (connectedWires[i] == null)
    			result++;
    	}
    	return result;
    }

    /**
     * @return Returns the port number of the next available port or -1,
     * if no more port is available
     */
    public final int getNextFreePort() {
    	for (int i = 0; i < getNumberOfTotalPorts(); i++) {
    		if (connectedWires[i] == null)
    			return i;
    	}
    	return -1;
    }
    
    /**
     * @return Returns an iterator with all connected wires
     */
	public final List getConnectedWires() {
		ArrayList temp = new ArrayList();
		for (int i = 0; i < getNumberOfTotalPorts(); i++) {
			temp.add(connectedWires[i]);
		}
	    return temp;
	}

	/**
	 * @param portNr The identifying port number
	 * @return Returns, wheter a port is free (no cable is plugged) or not free (cable is plugged)
	 */
	public final boolean isPortFree(int portNr) {
		return (connectedWires[portNr] == null);
	}
	
	/**
	 * @param portNr The identifying port number
	 * @return Returns a reference to the wire which is plugged in a specified port, or
	 * null, if no cable is currently plugged in this port
	 */
	public final Wire getWireInPort(int portNr) {
		return connectedWires[portNr];
	}
	
	/**
	 * @param wire The cable
	 * @return Returns the (first) port number, to which a cable is connected or
	 * -1, if cable is not connected at all
	 */
	public final int getPortOfWire(Wire wire) {
		for (int i = 0; i < getNumberOfTotalPorts(); i++) {
			if (connectedWires[i].equals(wire))
				return i;
		}
		return -1;
	}
	
	/**
	 * Checks, if a specified wire is connected or not 
	 * @param wireToTest Wire to test
	 * @return Returns, if wire is connected
	 */
	public final boolean isWireConnected(Wire wireToTest) {
	    if (wireToTest == null)
	        return false;
	    for (int i = 0; i < getNumberOfTotalPorts(); i++) {
	    	if (connectedWires[i].equals(wireToTest))
	    		return true;
	    }
	    return false;
	}

	/**
	 * This method must be overriden in subclasses. The configuration for the incoming network stack
	 * must be returned.
	 * @return The incoming network stack configuration
	 */
    protected abstract IncomingNetworkStackConfiguration getIncomingNetworkStackConfiguration();
    
    /**
     * This method must be overriden in subclasses. The configuration for the outgoing network stack
     * must be returned.
     * @return The outgoing network stack configuration
     */
    protected abstract OutgoingNetworkStackConfiguration getOutgoingNetworkStackConfiguration();
    	
	/*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.layer1.wire.WireConnector#bePlugged(at.jku.fim.datalinksimulation.layer1.wire.Wire,
	 *      int)
	 */
	public final void bePlugged(Wire wire, int portNr) throws WiringException {
		if (connectedWires[portNr] != null && !connectedWires[portNr].equals(wire))
			throw new WiringException("Port " + portNr + " of " + this.toString() + " is already occupied by another cable.");
		connectedWires[portNr] = wire;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.layer1.wire.WireConnector#beUnplugged(at.jku.fim.datalinksimulation.layer1.wire.Wire,
	 *      int)
	 */
	public final void beUnplugged(Wire wire, int portNr) throws WiringException {
		if (connectedWires[portNr] == null || !connectedWires[portNr].equals(wire))
			throw new WiringException("Port " + portNr + " of " + this.toString() + " is not connected to that cable.");
		connectedWires[portNr] = null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.layer1.wire.WireConnector#informAboutDataOnWire(at.jku.fim.datalinksimulation.layer1.wire.Wire,
	 *      at.jku.fim.datalinksimulation.data.RawData, java.util.Vector)
	 */
	public final void informAboutDataOnWire(Wire sender, RawData data) {
	    logger.log("Data received on Port <" + getPortOfWire(sender) + "> Wire <" + sender.toString() + ">");
	
	    networkStack.receiveFromWorld(data, getPortOfWire(sender));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see at.jku.fim.datalinksimulation.genericdevices.HardwareComponent#toString()
	 */
	public String toString() {
		return identifier + " - " + super.toString();
	}

	
	
}