package at.jku.fim.datalinksimulation.layer1.data;

import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * <p>Title: OSI Simulation Framework</p>
 * <p>Copyright: (c) 2004-05</p>
 * <p>Company: FIM@JKU, www.fim.uni-linz.ac.at</p>
 * @author Andreas Putzinger
 * @version $Id: RawData.java,v 1.1 2005/10/27 13:10:11 aputzinger Exp $
 */
public class RawData implements Cloneable {

    /**
     * contained data; only values between 0 and 255 are valid; if any other value is specified, it is set to 0
     */
    protected int[] rawData;

    // for iterator implementation - must be changed each time the rawData is modified
    protected int modCounter;

    /**
     * Constructor
     * @param data Inital content
     */
    public RawData(int[] data) {
        setRawDataBulk(data);
    }

    /**
     * Returns an iterator for the contained data.<BR>Content is returned Integer per Integer<BR>
     * Like in every Iterator implementation, the data is only valid as long as the data is not changed!
     * @return Returns a corresponding iterator
     */
    public Iterator iterator() {
        return new RawDataIterator();
    }

    /**
     * @return Returns the number of data bytes; 0 if empty
     */
    public int length() {
        if (rawData != null) {
            return rawData.length;
        }
        return 0;
    }

    /**
     * Set the raw data
     * @param data New raw data
     */
    public void setRawDataBulk(int[] data) {
        modCounter++;
        if (data == null)
            this.rawData = new int[0];
        else {
            this.rawData = (int[]) data.clone();
            for (int i = 0; i < data.length; i++) {
                if (rawData[i] > 255 || rawData[i] < 0)
                    rawData[i] = 0;
            }
        }
    }

    /**
     * @return Returns a copy of the raw data int array;<BR>
     * If no data is contained, an empty array is returned, but never null
     */
    public int[] getRawDataBulk() {
        if (rawData == null)
            return new int[0];
        return (int[]) rawData.clone();
    }

    /*
     *  (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
        if (rawData == null)
            return "";
        StringBuffer result = new StringBuffer();
        for (int i = 0; i < rawData.length; i++)
            result.append((char)rawData[i]);
        return result.toString();
    }

    /**
     * @return Returns printable string representation of string
     */
    public String stringRepresentation() {
        if (rawData == null)
            return "<null>";
        StringBuffer result = new StringBuffer();
        for (int i = 0; i < rawData.length; i++) {
            if (rawData[i] > 31)
                result.append((char)rawData[i]);
            else
                result.append("~");
        }
        return result.toString();
    } 
    
    /*
     *  (non-Javadoc) Deep-copy
     * @see java.lang.Object#clone()
     */
    public Object clone() {
        RawData myClone = null;
        try {
            myClone = (RawData) super.clone();
        } catch (CloneNotSupportedException ex) {
            // ignore
        }
        if (rawData != null) {
            int[] rawDataClone = (int[]) rawData.clone();
            myClone.setRawDataBulk(rawDataClone);
        }
        return myClone;
    }

    /**
     * Iterator Implementation
     * <p>Title: OSI Simulation Framework</p>
     * <p>Copyright: (c) 2004-05</p>
     * <p>Company: FIM@JKU, www.fim.uni-linz.ac.at</p>
     * @author Andreas Putzinger
     * @version Innerclass of $Id: RawData.java,v 1.1 2005/10/27 13:10:11 aputzinger Exp $
     */
    private class RawDataIterator implements Iterator {

        // position in array next to deliver; start with first byte at position 0
        private int posCounter = 0;

        // modification counter when instance was created; must always match to the counter w
        private int expectedModCounter;

        /**
         * Creates new instance of iterator; sets local modification counter
         */
        public RawDataIterator() {
            this.expectedModCounter = modCounter;
        }

        /**
         * Method not Implemented; Throws an <b>UnsupportedOperationException</b>!
         * @see java.util.Iterator#remove()
         */
        public void remove() {
            throw new UnsupportedOperationException(
                    "The method \"remove()\" is not supported by RawDataIterator implementation.");
        }

        /* (non-Javadoc)
         * @see java.util.Iterator#hasNext()
         */
        public boolean hasNext() {
            return (posCounter < rawData.length);
        }

        /**
         * Returnes next Integer object;<BR>
         * Throws ConcurrentModificationException if data was modified since creation of iterator<BR>
         * Throws NoSuchElementException if no Byte left
         * @see java.util.Iterator#next()
         */
        public Object next() {
            if (expectedModCounter != modCounter)
                throw new ConcurrentModificationException(
                        "Data was changed while iterating through - illegal iterator state.");
            if (!hasNext())
                throw new NoSuchElementException(
                        "There are no more elements left.");
            Integer result = new Integer(rawData[posCounter]);
            posCounter++;
            return result;
        }

    }

}