package at.jku.fim.datalinksimulation.layer2.protocols;

import at.jku.fim.datalinksimulation.layer1.data.RawData;
import at.jku.fim.datalinksimulation.utils.Logger;

/**
 * <p>Title: OSI Simulation Framework</p>
 * <p>Copyright: (c) 2004-05</p>
 * <p>Company: FIM@JKU, www.fim.uni-linz.ac.at</p>
 * @author Andreas Putzinger
 * @version $Id: EasyL2Utils.java,v 1.1 2005/10/27 13:10:11 aputzinger Exp $
 */

public class EasyL2Utils {

    /**
     * Interpret raw data as EasyL2 frame; if not successful, null is returned
     * @param data Data to interpret
     * @param logger logger instance; may not be null
     * @return Returns the interpreted EasyL2 frame or null
     */
    public static EasyL2FrameData interpretRawData(RawData data,
            Logger logger) {
        EasyL2FrameData content = new EasyL2FrameData();
        if (data == null) {
            logger.log("Cannot interpret EasyL2 frame - no data");
            return null;
        }
        // an EasyL2 frame must be at least 10 bytes long per definition (if no L3 data part contained)
        if (data.getRawDataBulk().length < 10) {
            logger.log("Cannot interpret EasyL2 frame - frame too short");
            return null;
        }

        int[] dataBulk = data.getRawDataBulk();
        
        // retrieve preamble at pos 0
        content.setPreamble(dataBulk[0]);
        
        // retrieve destination address at pos 1 and 2
        content.setDestination((dataBulk[1] << 8) + dataBulk[2]);
        
        // retrieve source address at pos 3 and 4
        content.setSource(((dataBulk[3]) << 8) + dataBulk[4]);
        
        // retrieve specified frame size at pos 5
        content.setSize(dataBulk[5]);
        
        // if specified frame size does not match with actual --> corrupt frame
        if (content.getSize() != dataBulk.length) {
            logger.log("Cannot interpret EasyL2 frame - specified frame size does not match with actual size");
            return null;
        }
        
        // retrieve layer 3 data starting at pos 6
        int [] l3Data = new int[dataBulk.length - 10];
        for (int i = 0; i < dataBulk.length - 10; i++) {
            l3Data[i] = dataBulk[6 + i];
        }
        content.setData(l3Data);
        
        // retrieve specified CRC
        int[] crc = new int[4];
        for (int i = 0; i < 4; i++) {
            crc[i] = dataBulk[dataBulk.length - 4 + i];
        }
        content.setCrc(crc);
        
        logger.log("EasyL2 frame interpretation successful - only few sanity checks done");
        return content;
    }

    /**
     * "Serialize" an EasyL2 object so that it can be transferred over network
     * @param content EasyL2 data
     * @param logger logger instance, which may NOT be null
     * @return Returns raw data of frame
     */
    public static RawData createFrame(EasyL2FrameData content, Logger logger) {
        RawData frame = new RawData(null);
        if (content != null) {
            int[] frameData = new int[content.getData().length + 10];
            
            // set magic number 0x7F; preamble
            frameData[0] = 0x7F;

            // set msb of destination address
            frameData[1] = (255 & (content.getDestination() >>> 8));
            
            // set lsb of destination address
            frameData[2] = (255 & (content.getDestination()));
            
            // set msb of source address
            frameData[3] = (255 & (content.getSource() >>> 8));
            
            // set lsb of source address
            frameData[4] = (255 & (content.getSource()));
            
            // set frame length
            frameData[5] = frameData.length;
            
            // insert layer 3 data
            for (int i = 0; i < content.getData().length; i++) {
                frameData[6 + i] = content.getData()[i];
            }
            
            // Insert CRC
            int[] crc = content.getCrc();
            if (crc == null || crc.length != 4)
                crc = new int[4];
            frameData[frameData.length - 4] = crc[0];
            frameData[frameData.length - 3] = crc[1];
            frameData[frameData.length - 2] = crc[2];
            frameData[frameData.length - 1] = crc[3];
            frame.setRawDataBulk(frameData);
        }
        return frame;
    }

}